@@ -6,7 +6,7 @@ use warnings;
 
 use Data::Dumper;
 
-our $VERSION = '0.09';
+our $VERSION = '0.10';
 
 =head1 NAME
 
@@ -14,7 +14,7 @@ Protocol::ACME - Interface to the Let's Encrypt ACME API
 
 =head1 VERSION
 
-Version 0.09
+Version 0.10
 
 =head1 SYNOPSIS
 
@@ -38,17 +38,13 @@ Version 0.09
                                  );
 
    $acme->directory();
-
    $acme->register();
-
    $acme->accept_tos();
 
    for my $domain ( @names )
    {
      $acme->authz( $domain );
-
      $acme->handle_challenge( $challenges->{$domain} );
-
      $acme->check_challenge();
    }
 
@@ -56,10 +52,10 @@ Version 0.09
  };
  if ( $@ )
  {
-   die $@ if ref $@ ne "Protocol::ACME::Exception";
-   print "Error occured: Status: $@->{status},
-                         Detail: $@->{detail},
-                         Type: $@->{type}\n";
+   die if !UNIVERSAL::isa($@, 'Protocol::ACME::Exception');
+   die "Error occured: Status: $@->{status},
+                       Detail: $@->{detail},
+                       Type:   $@->{type}\n";
  }
  else
  {
@@ -100,6 +96,7 @@ a * are required.
    ua                      HTTP::Tiny->new()
    loglevel                error
    debug                   0
+   mailto                  undef
 
 B<host>: The API end point to connect to.  This will generally be acme-staging.api.letsencrypt.org
 or acme-v01.api.letsencrypt.org
@@ -116,6 +113,9 @@ B<loglevel>: Set the loglevel to one of the C<Log::Any> values.
 
 B<debug>: If set to non-zero this is a shortcut for C<loglevel => debug>
 
+B<mailto>: This should be the email address that you want associated with your account.  This is used
+my Let's Encrypt for expiration notification.
+
 =back
 
 =head2 METHODS
@@ -166,12 +166,22 @@ Loads the directory from the ACME host.  This call must be made first
 before any other calls to the API in order the bootstrap the API
 resource list.
 
-=item register()
+=item register( %args )
 
 Call the new-reg resource and create an account associated with the
 loaded account key.  If that key has already been registered this method
 will gracefully and silently handle that.
 
+Arguments that can be passed in:
+
+   KEY                     DEFAULT
+   -----------             --------------------
+   mailto                  undef
+
+B<mailto>: See C<new> for a desciption.  This will override the value passed to new
+if any.
+
+
 =item accept_tos()
 
 In order to use the Let's Encrypt service, the account needs to accept
@@ -260,6 +270,10 @@ try and determine what the format it.
 
 On success C<Protocol::ACME> will return the DER encoded signed certificate.
 
+=item $cert_chain = chain()
+
+After C<sign> has been called and a cert successfully created, C<chain> will
+fetch and return the DER encoded certificate issuer.
 
 =item revoke( $certfile )
 
@@ -276,93 +290,12 @@ called.
 
 =cut
 
-
-package Log::Any::Adapter::AcmeLocal;
-
-use Log::Any::Adapter::Util ();
-use Time::HiRes qw( gettimeofday );
-use base qw/Log::Any::Adapter::Base/;
-
-my $trace_level = Log::Any::Adapter::Util::numeric_level('trace');
-
-sub init {
-    my ($self) = @_;
-    if ( exists $self->{log_level} ) {
-        $self->{log_level} =
-          Log::Any::Adapter::Util::numeric_level( $self->{log_level} )
-          unless $self->{log_level} =~ /^\d+$/;
-    }
-    else {
-        $self->{log_level} = $trace_level;
-    }
-}
-
-foreach my $method ( Log::Any::Adapter::Util::logging_methods() ) {
-    no strict 'refs';
-    my $method_level = Log::Any::Adapter::Util::numeric_level($method);
-    *{$method} = sub {
-        my ( $self, $text ) = @_;
-        return if $method_level > $self->{log_level};
-
-        my ( $sec, $usec ) = gettimeofday();
-
-        printf STDOUT "%d.%06d %s\n", $sec, $usec, $text;
-    };
-}
-
-foreach my $method ( Log::Any::Adapter::Util::detection_methods() ) {
-    no strict 'refs';
-    my $base = substr( $method, 3 );
-    my $method_level = Log::Any::Adapter::Util::numeric_level($base);
-    *{$method} = sub {
-        return !!( $method_level <= $_[0]->{log_level} );
-    };
-}
-
-
-package Protocol::ACME::Exception;
-
-use Data::Dumper;
-
-# very simple stringification ... make this
-# more elaborate according to taste
-use overload ('""' => \&stringify);
-sub stringify
-{
-    my $self = shift;
-    return ref($self).' error: '.Dumper $self;
-}
-
-sub new
-{
-  my $class = shift;
-
-  my $error = shift;
-  my $self = { status => 0, detail => "", type => "unknown" };
-
-  if ( ref $error eq "HASH" )
-  {
-    @$self{keys %$error} = values %$error;
-  }
-  elsif ( ref $error )
-  {
-    $self->{detail} = "double error: bad arg passed to exception constructor";
-  }
-  else
-  {
-    $self->{detail} = $error;
-  }
-
-  bless $self, $class;
-
-  return $self;
-}
-
 package Protocol::ACME;
 
 use strict;
 use warnings;
 
+use Protocol::ACME::Exception;
 use Protocol::ACME::Utils;
 
 use Crypt::Format;
@@ -373,9 +306,6 @@ use MIME::Base64 qw( encode_base64url decode_base64url decode_base64 encode_base
 use HTTP::Tiny;
 use JSON;
 use Digest::SHA qw( sha256 );
-use Log::Any;
-use Log::Any::Adapter;
-
 use Carp;
 
 
@@ -413,14 +343,13 @@ sub _init
   $self->{openssl}  = $args->{openssl} if exists $args->{openssl};
   $self->{debug}    = $args->{debug}   if exists $args->{debug};
   $self->{loglevel} = exists $args->{loglevel} ? $args->{loglevel} : "error";
+  $self->{contact}->{mailto} = $args->{mailto} if exists $args->{mailto};
 
   if ( $self->{debug} )
   {
     $self->{loglevel} = "debug";
   }
 
-  print $self->{loglevel}, "\n";
-
   if ( ! exists $self->{ua} )
   {
     $self->{ua} = HTTP::Tiny->new( agent => $USERAGENT, verify_SSL => 1 );
@@ -431,8 +360,11 @@ sub _init
     _throw( detail => "host parameter is required for Protocol::ACME::new" );
   }
 
-  $self->{log} = Log::Any->get_logger();
-  Log::Any::Adapter->set('AcmeLocal', log_level => $self->{loglevel} );
+  $self->{log} = $args->{'logger'} || do
+  {
+    require Protocol::ACME::Logger;
+    Protocol::ACME::Logger->new($self->{loglevel});
+  };
 
   if ( exists $args->{account_key} )
   {
@@ -489,7 +421,7 @@ sub account_key
   }
   else
   {
-    map { $args{$_} = $key->{$_} } keys %$key;
+    @args{ keys %$key } = values %$key;
   }
 
   if ( $args{filename} )
@@ -497,7 +429,7 @@ sub account_key
     $args{buffer} = _slurp( $args{filename} );
     if ( ! $args{buffer} )
     {
-      _throw( "Could not load key from file $args{filename}" );
+      _throw( "Could not load key from file $args{filename}: $!" );
     }
   }
 
@@ -543,7 +475,7 @@ sub account_key
 
   if ( ! $key )
   {
-    die Protocol::ACME::Exception->new( { detail => "Could not load key into key structure" } );
+    _throw( "Could not load key into key structure" );
   }
 
   $key->use_sha256_hash();
@@ -571,10 +503,10 @@ sub directory
 
   if ( $resp->{status} != 200 )
   {
-    die Protocol::ACME::Exception->new( { detail => "Failed to fetch the directory for $self->{host}", resp => $resp } );
+    _throw( detail => "Failed to fetch the directory for $self->{host}", resp => $resp );
   }
 
-  my $data = decode_json( $resp->{content} );
+  my $data = _decode_json( $resp->{content} );
 
   @{$self->{links}}{keys %$data} = values %$data;
 
@@ -588,8 +520,22 @@ sub directory
 sub register
 {
   my $self = shift;
+  my %args = @_;
+
+  my $obj = {};
+  $obj->{resource} = 'new-reg';
+
+  if ( exists $args{mailto} )
+  {
+    push @{$obj->{contact}}, $args{mailto};
+  }
+  elsif ( exists $self->{contact}->{mailto} )
+  {
+    push @{$obj->{contact}}, $self->{contact}->{mailto};
+  }
+
+  my $msg = _encode_json( $obj );
 
-  my $msg = encode_json( { resource => 'new-reg' } );
   my $json = $self->_create_jws( $msg );
 
   $self->{log}->debug( "Sending registration message" );
@@ -603,7 +549,7 @@ sub register
     $self->{log}->debug( "Known key used" );
     $self->{log}->debug( "Refetching with location URL" );
 
-    my $json = $self->_create_jws( encode_json( { "resource" => 'reg' } ) );
+    my $json = $self->_create_jws( _encode_json( { "resource" => 'reg' } ) );
 
     $resp = $self->_request_post( $self->{links}->{'reg'}, $json );
 
@@ -615,7 +561,7 @@ sub register
     }
     else
     {
-      die Protocol::ACME::Exception->new( $self->{content} );
+      _throw( @{ $self->{content} } );
     }
   }
   elsif ( $resp->{status} == 201 )
@@ -629,7 +575,7 @@ sub register
   }
   else
   {
-    die Protocol::ACME::Exception->new( $self->{content} );
+    _throw( @{ $self->{content} } );
   }
 
 
@@ -649,7 +595,7 @@ sub recovery_key
 
 
   my $pem = _slurp( $keyfile );
-  croak( Protocol::ACME::Exception->new( { detail => "$keyfile: $!" } ) ) if ! $pem;
+  _throw( "$keyfile: $!" ) if ! $pem;
 
   my $url = "https://acme-staging.api.letsencrypt.org/acme/reg/101834";
 
@@ -671,7 +617,7 @@ sub recovery_key
               }
             };
 
-  my $json = $self->_create_jws( _hash_to_json($msg) );
+  my $json = $self->_create_jws( _encode_json($msg) );
 
   my $resp = $self->_request_post( $url, $json );
 
@@ -692,7 +638,7 @@ sub accept_tos
   # TODO: check for existance of terms-of-service link
   # TODO: assert on reg url being present
 
-  my $msg = _hash_to_json( { "resource"  => "reg",
+  my $msg = _encode_json( { "resource"  => "reg",
                              "agreement" => $self->{links}->{'terms-of-service'},
                              "key"       => { "e"   => $self->{key}->{e},
                                              "kty" => "RSA",
@@ -709,7 +655,7 @@ sub accept_tos
   }
   else
   {
-    die Protocol::ACME::Exception->new( $self->{content} );
+    _throw( @{ $self->{content} } );
   }
 }
 
@@ -724,12 +670,11 @@ sub revoke
 
   if ( ! $cert )
   {
-    # TODO: should be an ACME exception
-    die "Could not load cert form $certfile";
+    _throw("Could not load cert from $certfile: $!");
   }
 
 
-  my $msg = _hash_to_json( { "resource"    => "revoke-cert",
+  my $msg = _encode_json( { "resource"    => "revoke-cert",
                             "certificate" => encode_base64url( $cert ) } );
 
 
@@ -739,7 +684,7 @@ sub revoke
 
   if ( $resp->{status} != 200 )
   {
-    die Protocol::ACME::Exception->new( $self->{content} );
+    _throw( @{ $self->{content} } );
   }
 
 }
@@ -752,7 +697,7 @@ sub authz
   $self->{log}->debug( "Sending authz message for $domain" );
   # TODO: check for 'next' URL and that is it authz
 
-  my $msg = _hash_to_json( { "identifier" => { "type" => "dns", "value" => $domain },
+  my $msg = _encode_json( { "identifier" => { "type" => "dns", "value" => $domain },
                             "resource"   => "new-authz" } );
 
   my $json = $self->_create_jws( $msg );
@@ -765,7 +710,7 @@ sub authz
   }
   else
   {
-    die Protocol::ACME::Exception->new( $self->{content} );
+    _throw( @{ $self->{content} } );
   }
 }
 
@@ -777,7 +722,7 @@ sub handle_challenge
 
   my $key = $self->{key};
 
-  my $jwk = _hash_to_json( { "e" => $key->{e}, "kty" => "RSA", "n" => $key->{n} } );
+  my $jwk = _encode_json( { "e" => $key->{e}, "kty" => "RSA", "n" => $key->{n} } );
   my $token;
   my $challenge_url;
 
@@ -808,7 +753,7 @@ sub handle_challenge
   }
   else
   {
-    die Protocol::ACME::Exception->new( { status => 0, detail => $ret, type => "challenge_exec" } );
+    _throw( status => 0, detail => $ret, type => "challenge_exec" );
   }
 }
 
@@ -817,7 +762,7 @@ sub check_challenge
 {
   my $self = shift;
 
-  my $msg = _hash_to_json( { "resource" => "challenge", "keyAuthorization" => $self->{token} . '.' . $self->{fingerprint} } );
+  my $msg = _encode_json( { "resource" => "challenge", "keyAuthorization" => $self->{token} . '.' . $self->{fingerprint} } );
 
   my $json = $self->_create_jws( $msg );
 
@@ -862,11 +807,11 @@ sub sign
   }
   elsif( ref $csr eq "SCALAR" )
   {
-    $args{Bbuffer} = $$csr;
+    $args{buffer} = $$csr;
   }
   else
   {
-    map { $args{$_} = $csr->{$_} } keys %$csr;
+    @args{keys %$csr} = values %$csr;
   }
 
   if ( $args{filename} )
@@ -890,20 +835,47 @@ sub sign
 
   my $der = $args{format} eq "DER" ? $args{buffer} : Crypt::Format::pem2der( $args{buffer} );
 
-  my $msg = _hash_to_json( { "resource" => "new-cert", "csr" => encode_base64url( $der ) } );
+  my $msg = _encode_json( { "resource" => "new-cert", "csr" => encode_base64url( $der ) } );
 
   my $json = $self->_create_jws( $msg );
 
-  my $resp = $self->_request_post( $self->{links}->{'new-cert'}, $json );
+  my $resp = $self->_request_post( $self->{links}->{'new-cert'}, $json, 1 );
 
   if ( $resp->{status} != 201 )
   {
-    die Protocol::ACME::Exception->new( $self->{content} );
+    _throw( %{_decode_json($resp->{content}) } );
   }
 
-  my $cert = $resp->{content};
+  my $links = _link_to_hash( $resp->{headers}->{'link'} );
 
-  return $cert;
+  $self->{links}->{chain} = $links->{up} if exists $links->{up};
+  $self->{links}->{cert}  = $resp->{headers}->{location} if exists $resp->{headers}->{location};
+
+  $self->{cert} = $resp->{content};
+
+  return $self->{cert};
+}
+
+sub chain
+{
+  my $self = shift;
+
+  if ( ! exists $self->{links}->{chain} )
+  {
+    _throw( "URL for the cert chain missing.  Has sign() been called yet?" );
+  }
+
+  my $resp = $self->_request_get( $self->{links}->{chain}, 1 );
+
+  if ( $resp->{status} != 200 )
+  {
+    _throw( detail => "Error received fetching the certificate chain",
+            status => $resp->{status}  );
+  }
+
+  $self->{chain} = $resp->{content};
+
+  return $self->{chain};
 }
 
 #############################################################
@@ -913,23 +885,28 @@ sub _request_get
 {
   my $self = shift;
   my $url  = shift;
+  my $nodecode = shift || 0;
 
   my $resp = $self->{ua}->get( $url );
 
   $self->{nonce} = $resp->{headers}->{$NONCE_HEADER};
   $self->{json} = $resp->{content};
 
-  eval {
-    $self->{content} = decode_json( $resp->{content} );
-  };
+  #Exception here should be fatal.
+  $self->{content} = undef;
+  $self->{content} = _decode_json( $resp->{content} ) unless $nodecode;
+
+  $self->{response} = $resp;
+
   return $resp;
 }
 
 sub _request_post
 {
-  my $self    = shift;
-  my $url     = shift;
-  my $content = shift;
+  my $self     = shift;
+  my $url      = shift;
+  my $content  = shift;
+  my $nodecode = shift || 0;
 
   my $resp = $self->{ua}->post( $url, { content => $content } );
 
@@ -937,9 +914,12 @@ sub _request_post
 
   $self->{json} = $resp->{content};
 
-  eval {
-    $self->{content} = decode_json( $resp->{content} );
-  };
+  #Let exception from decode_json() propagate:
+  #if we failed to decode the JSON, that’s a show-stopper.
+  $self->{content} = undef;
+  $self->{content} = _decode_json( $resp->{content} ) unless $nodecode;
+
+  $self->{response} = $resp;
 
   return $resp;
 }
@@ -960,18 +940,9 @@ sub _slurp
 {
   my $filename = shift;
 
-  my $fh = IO::File->new( $filename );
-  if ( ! $fh )
-  {
-    return;
-  }
+  open my $fh, '<', $filename or return undef;
 
-  my $content;
-
-  while( <$fh> )
-  {
-    $content .= $_;
-  }
+  sysread( $fh, my $content, -s $fh ) or return undef;
 
   return $content;
 }
@@ -982,6 +953,13 @@ sub _link_to_hash
   my $arrayref = shift;
   my $links;
 
+  return {} unless $arrayref;
+
+  if ( ! ref $arrayref )
+  {
+    $arrayref = [ $arrayref ];
+  }
+
   for my $link ( @$arrayref )
   {
     my ( $value, $key ) = split( ';', $link );
@@ -1001,33 +979,6 @@ sub _link_to_hash
   return $links;
 }
 
-sub _hash_to_json
-{
-  my $hash = shift;
-  my $json = "{";
-  my $quote = '"';
-  my $colon = ':';
-  my $comma = ',';
-
-  for ( sort keys %$hash )
-  {
-    # die "hash_to_json does not handle nested references yet" if ref $hash->{$_};
-    if ( ref $hash->{$_} eq "HASH" )
-    {
-      $json .= $quote . $_ . $quote . $colon . _hash_to_json($hash->{$_}) . $comma;
-    }
-    else
-    {
-      $json .= $quote . $_ . $quote . $colon . $quote . $hash->{$_} . $quote . $comma;
-    }
-  }
-
-  $json =~ s/,$//;
-
-  $json .= '}';
-
-}
-
 sub _bigint_to_binary {
     my ( $bigint ) = @_;
 
@@ -1065,12 +1016,27 @@ sub _create_jws_internal
               payload   => encode_base64url( $msg ),
               signature => $sig };
 
-  my $json = _hash_to_json( $jws );
+  my $json = _encode_json( $jws );
 
   return $json;
 
 }
 
+sub _decode_json
+{
+  my $ref = shift;
+  return JSON->new->allow_nonref->decode($ref);
+}
+
+sub _encode_json
+{
+  my $ref = shift;
+#  my $json = JSON->new();
+#  $json->canonical();
+  #  return $json->encode($ref);
+  return JSON->new->canonical->encode($ref);
+}
+
 
 =head1 AUTHOR
 
